package ru.dbotthepony.mc.otm.datagen.tags

import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.core.Registry
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.BlockTags
import net.minecraft.tags.GameEventTags
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraft.world.item.Tier
import net.minecraft.world.item.Tiers
import net.minecraft.world.level.block.Block
import net.minecraftforge.data.event.GatherDataEvent
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.datagen.DataGen
import java.util.stream.Stream
import net.minecraft.data.tags.TagsProvider as MinecraftTagsProvider

interface MTagAppender<T> {
	fun add(value: T): MTagAppender<T>
	fun addSafe(value: T): Boolean
	fun add(vararg values: T): MTagAppender<T>
	fun addSafe(vararg values: T): Boolean {
		var any = false
		for (value in values) any = addSafe(value)
		return any
	}

	fun add(values: Collection<T>): MTagAppender<T>
	fun add(values: Stream<out T>): MTagAppender<T> = add(values.toList())

	fun addSafe(values: Collection<T>): Boolean {
		var any = false
		for (value in values) any = addSafe(value)
		return any
	}
}

interface ForgeTagAppender<T> : MTagAppender<T> {
	fun add(key: String, value: T): MTagAppender<T>
	fun addSafe(key: String, value: T): Boolean
	fun add(key: String, vararg values: T): MTagAppender<T>
	fun addSafe(key: String, vararg values: T): Boolean {
		var any = false
		for (value in values) any = addSafe(key, value)
		return any
	}

	fun add(key: String, values: Collection<T>): MTagAppender<T>

	fun addSafe(key: String, values: Collection<T>): Boolean {
		var any = false
		for (value in values) any = addSafe(key, value)
		return any
	}
}

class TagsProvider(
	private val event: GatherDataEvent
) {
	inner class Delegate<T>(registry: Registry<T>) : MinecraftTagsProvider<T>(event.generator, registry, DataGen.MOD_ID, event.existingFileHelper) {
		init {
			event.generator.addProvider(true, this)
		}

		private val tags = HashMap<TagKey<T>, ObjectArraySet<T>>()

		override fun addTags() {
			if (tags.isEmpty()) {
				return
			}

			for ((tag, values) in tags) {
				tag(tag).also {
					for (value in values) {
						it.add(value)
					}
				}
			}
		}

		fun getSet(tag: TagKey<T>): MutableSet<T> {
			return tags.computeIfAbsent(tag) { ObjectArraySet() }
		}

		fun getSet(location: ResourceLocation) = getSet(TagKey.create(registry.key(), location))

		fun appender(tag: TagKey<T>, message: ((T) -> Any) = { "$it is already in $tag" }): MTagAppender<T> {
			val list = getSet(tag)

			return object : MTagAppender<T> {
				override fun add(value: T): MTagAppender<T> {
					if (!list.add(value)) {
						throw IllegalStateException(message.invoke(value).toString())
					}

					return this
				}

				override fun add(values: Collection<T>): MTagAppender<T> {
					for (value in values) add(value)
					return this
				}

				override fun addSafe(value: T): Boolean {
					return list.add(value)
				}

				override fun add(vararg values: T): MTagAppender<T> {
					values.forEach(this::add)
					return this
				}
			}
		}

		fun forge(path: String): ForgeTagAppender<T> {
			val parent by lazy { appender(ResourceLocation("forge", path)) }

			return object : ForgeTagAppender<T> {
				override fun add(key: String, value: T): MTagAppender<T> {
					val tag = TagKey.create(registry.key(), ResourceLocation("forge", "$path/$key"))

					if (!getSet(tag).add(value)) {
						throw IllegalStateException("$value is already in $tag")
					}

					return this
				}

				override fun addSafe(key: String, value: T): Boolean {
					val tag = TagKey.create(registry.key(), ResourceLocation("forge", "$path/$key"))
					return getSet(tag).add(value)
				}

				override fun add(key: String, vararg values: T): MTagAppender<T> {
					for (value in values) add(key, value)
					return this
				}

				override fun add(key: String, values: Collection<T>): MTagAppender<T> {
					for (value in values) add(key, value)
					return this
				}

				override fun add(value: T): MTagAppender<T> {
					return parent.add(value)
				}

				override fun addSafe(value: T): Boolean {
					return parent.addSafe(value)
				}

				override fun add(vararg values: T): MTagAppender<T> {
					return parent.add(*values)
				}

				override fun add(values: Collection<T>): MTagAppender<T> {
					return parent.add(values)
				}
			}
		}

		fun appender(location: ResourceLocation): MTagAppender<T> {
			return appender(TagKey.create(registry.key(), location))
		}

		fun appender(location: ResourceLocation, message: (T) -> Any): MTagAppender<T> {
			return appender(TagKey.create(registry.key(), location), message)
		}
	}

	val blocks = Delegate(Registry.BLOCK)
	val items = Delegate(Registry.ITEM)
	val mobEffects = Delegate(Registry.MOB_EFFECT)

	val androidImmuneEffects = mobEffects.appender(MatteryPlayerCapability.ANDROID_IMMUNE_EFFECTS)

	val requiresShovel = blocks.appender(BlockTags.MINEABLE_WITH_SHOVEL)
	val requiresAxe = blocks.appender(BlockTags.MINEABLE_WITH_AXE)
	val requiresHoe = blocks.appender(BlockTags.MINEABLE_WITH_HOE)
	val requiresPickaxe = blocks.appender(BlockTags.MINEABLE_WITH_PICKAXE)

	val requiresStoneTool = blocks.appender(BlockTags.NEEDS_STONE_TOOL)
	val requiresIronTool = blocks.appender(BlockTags.NEEDS_IRON_TOOL)
	val requiresDiamondTool = blocks.appender(BlockTags.NEEDS_DIAMOND_TOOL)

	val witherImmune = blocks.appender(BlockTags.WITHER_IMMUNE)
	val dragonImmune = blocks.appender(BlockTags.DRAGON_IMMUNE)

	fun ore(key: String, block: Block): TagsProvider {
		val forgeKey = ResourceLocation("forge", "ores/$key")
		val b = TagKey.create(Registry.BLOCK_REGISTRY, forgeKey)
		val i = TagKey.create(Registry.ITEM_REGISTRY, forgeKey)

		items.getSet(i).add(block.asItem())
		itemOres.add(block.asItem())
		blocks.getSet(b).add(block)
		blockOres.add(block)

		return this
	}

	fun clump(key: String, clump: Item): TagsProvider {
		rawMaterials.add(key, clump)
		rawMaterials.add(clump)
		return this
	}

	fun clump(key: String, clump: Item, rawBlock: Block): TagsProvider {
		val rawKey = "raw_$key"

		rawMaterials.add(key, clump)
		rawMaterials.add(clump)
		storageBlocksAsItem.add(rawKey, rawBlock.asItem())
		storageBlocksAsItem.add(rawBlock.asItem())
		storageBlocksAsBlock.add(rawKey, rawBlock)
		storageBlocksAsBlock.add(rawBlock)

		return this
	}

	val circuits = items.forge("circuits")
	val dusts = items.forge("dusts")
	val ingots = items.forge("ingots")
	val itemOres = items.forge("ores")
	val blockOres = blocks.forge("ores")
	val plates = items.forge("plates")
	val storageBlocksAsItem = items.forge("storage_blocks")
	val storageBlocksAsBlock = blocks.forge("storage_blocks")
	val rawMaterials = items.forge("raw_materials")
	val wires = items.forge("wires")

	val gameEvents = Delegate(Registry.GAME_EVENT)
	val vibrations = gameEvents.appender(GameEventTags.VIBRATIONS)

	fun requiresPickaxe(block: Block, tier: Tier? = null): TagsProvider {
		requiresPickaxe.add(block)

		when (tier) {
			Tiers.STONE -> requiresStoneTool.add(block)
			Tiers.IRON -> requiresIronTool.add(block)
			Tiers.DIAMOND, Tiers.NETHERITE -> requiresDiamondTool.add(block)
		}

		return this
	}

	fun requiresPickaxe(blocks: Collection<Block>, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresPickaxe(block, tier)
		return this
	}

	fun requiresPickaxe(vararg blocks: Block, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresPickaxe(block, tier)
		return this
	}

	fun requiresShovel(block: Block, tier: Tier? = null): TagsProvider {
		requiresShovel.add(block)

		when (tier) {
			Tiers.STONE -> requiresStoneTool.add(block)
			Tiers.IRON -> requiresIronTool.add(block)
			Tiers.DIAMOND, Tiers.NETHERITE -> requiresDiamondTool.add(block)
		}

		return this
	}

	fun requiresShovel(blocks: Collection<Block>, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresShovel(block, tier)
		return this
	}

	fun requiresAxe(block: Block, tier: Tier? = null): TagsProvider {
		requiresAxe.add(block)

		when (tier) {
			Tiers.STONE -> requiresStoneTool.add(block)
			Tiers.IRON -> requiresIronTool.add(block)
			Tiers.DIAMOND, Tiers.NETHERITE -> requiresDiamondTool.add(block)
		}

		return this
	}

	fun requiresAxe(blocks: Collection<Block>, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresAxe(block, tier)
		return this
	}
}

