package ru.dbotthepony.mc.otm.matter

import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.ItemTags
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.set

sealed class AbstractRegistryAction : Comparable<AbstractRegistryAction> {
	sealed class Condition {
		abstract fun test(): Boolean
	}

	class CombinedCondition(val conditions: Collection<Condition>) : Condition() {
		override fun test(): Boolean {
			return conditions.all { it.test() }
		}
	}

	val errorOnFailure: Boolean
	val tag: TagKey<Item>?
	val key: ResourceLocation?
	val matter: ImpreciseFraction?
	val complexity: Double?
	val priority: Int?

	constructor(json: JsonObject) {
		errorOnFailure = json["error_on_failure"]?.asBoolean ?: false
		priority = json["priority"]?.asInt

		val id = json["id"]?.asString ?: throw JsonParseException("Missing `id` value")

		if (id.startsWith("#")) {
			if (id.startsWith("#:")) {
				throw JsonSyntaxException("Invalid `id` value: $id")
			}

			tag = ItemTags.create(
				ResourceLocation.tryParse(id.substring(1)) ?: throw JsonSyntaxException("Invalid `id` value: $id")
			)
			key = null
		} else {
			key = ResourceLocation.tryParse(id) ?: throw JsonSyntaxException("Invalid `id` value: $id")
			tag = null
		}

		try {
			matter = json["matter"]?.asString?.let(::ImpreciseFraction)

			if (matter != null && !matter.isPositive) {
				throw JsonParseException("Can't have non-positive matter value. To remove an entry from registry please use 'delete' action instead.")
			}
		} catch(err: NumberFormatException) {
			throw JsonParseException("Invalid `matter` field: ${json["matter"]}", err)
		}

		try {
			complexity = json["complexity"]?.asString?.toDouble()

			if (complexity != null && complexity <= 0.0) {
				throw JsonParseException("Can't have non-positive complexity. To remove an entry from registry please use 'delete' action instead.")
			}
		} catch(err: NumberFormatException) {
			throw JsonParseException("Invalid `complexity` field: ${json["complexity"]}", err)
		}
	}

	constructor(
		tag: TagKey<Item>,
		matter: ImpreciseFraction?,
		complexity: Double?,
		priority: Int? = null,
		errorOnFailure: Boolean = false
	) {
		this.tag = tag
		this.key = null
		this.matter = matter
		this.complexity = complexity
		this.priority = priority
		this.errorOnFailure = errorOnFailure

		if (matter != null && !matter.isPositive) {
			throw IllegalArgumentException("Can't have non-positive matter value. To remove an entry from registry please use 'delete' action instead.")
		}

		if (complexity != null && complexity <= 0.0) {
			throw IllegalArgumentException("Can't have non-positive complexity. To remove an entry from registry please use 'delete' action instead.")
		}
	}

	constructor(
		key: ResourceLocation,
		matter: ImpreciseFraction?,
		complexity: Double?,
		priority: Int? = null,
		errorOnFailure: Boolean = false
	) {
		this.tag = null
		this.key = key
		this.matter = matter
		this.complexity = complexity
		this.priority = priority
		this.errorOnFailure = errorOnFailure

		if (matter != null && !matter.isPositive) {
			throw IllegalArgumentException("Can't have non-positive matter value. To remove an entry from registry please use 'delete' action instead.")
		}

		if (complexity != null && complexity <= 0.0) {
			throw IllegalArgumentException("Can't have non-positive complexity. To remove an entry from registry please use 'delete' action instead.")
		}
	}

	fun checkConditions(): Boolean {
		return true
	}

	internal abstract fun update(registry: MutableCollection<MutableEntry>, modifier: ResourceLocation)

	protected inline fun fail(reason: () -> String) {
		if (errorOnFailure) {
			throw JsonParseException(reason.invoke())
		}
	}

	fun canReplace(other: AbstractRegistryAction?): Boolean {
		if (other == null)
			return true

		if (other.priority != null && priority != null)
			return other.priority <= priority

		return other.priority == null
	}

	override fun compareTo(other: AbstractRegistryAction): Int {
		if (other.priority == null) {
			if (priority == null) {
				return 0
			}

			return 1
		}

		if (priority == null) {
			return -1
		}

		return priority.compareTo(other.priority)
	}

	open fun toJson(): JsonObject {
		return JsonObject().also {
			if (key != null)
				it["id"] = JsonPrimitive(key.toString())
			else
				it["id"] = JsonPrimitive("#${tag!!.location}")

			if (priority != null)
				it["priority"] = priority

			if (matter != null)
				it["matter"] = matter.toString()

			if (complexity != null)
				it["complexity"] = complexity
		}
	}
}
